DOMAIN = 'gene_sets'
GENE_SET_ATTRIBUTES = ('gs_id', 'hierarchy', 'organism', 'name', 'genes', 'description', 'link')



# cytoband
CYTOBAND_DOWNLOAD_LINK = 'http://statweb.stanford.edu/~tibs/GSA/cytobands-stanford.gmt'

# Reactome
REACTOME_DOWNLOAD_LINK = 'http://www.reactome.org/download/current/ReactomePathways.gmt.zip'
REACTOME_FILE_NAME = 'ReactomePathways.gmt'
REACTOME_DETAIL_LINK = 'https://reactome.org/content/detail/{}'

# GO
GO_TERM_LINK = 'http://amigo.geneontology.org/amigo/term/{}'


# OMIM
OMIM_LINK = 'http://www.omim.org/entry/{}'
